/*
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include "fontTT.h"

static FT_Library  Library;

typedef struct _IFT_FontRec {
  char * name;
  char * file;
  char * fullpath;
} IFT_FontRec;

static IFT_FontRec * FontList=NULL;
static int FontListNum=0;

char IFT_Error[1024];

static void IFT_SetError(char *fmt, ...); 

static int _IFT_FontListNew(char **dirnames);

static void _IFT_FontListFree(void);

static int _IFT_FontLookup(char *dirname); 
static int _IFT_FontAdd(char * file, char * fullpath);
static int _IFT_FontRecCmpName(const void *p1, const void *p2); 
static int _IFT_FontRecCmpFile(const void *p1, const void *p2); 
static char * _IFT_GetFontFile(char * name,char *style);
static char * _IFT_SearchFontFile(char * name, char * file);
static void _IFT_Blit(IFT_Pixbuf* buf, FT_Bitmap bitmap, int x0, int y0,
                      unsigned char *col);
static void _IFT_CopyPixbuf(IFT_Pixbuf * dstbuf, int dstx, int dsty, 
                            IFT_Pixbuf * srcbuf, int srcx, int srcy, int srcw,
                            int srch);
static int _IFT_LoadGlyphs (IFT * ift);
static int _IFT_RenderToPixbuf(IFT *ift);

#define affect_if_min(var,val) { if ((val) < var) { var=(val); } }
#define affect_if_max(var,val) { if ((val) > var) { var=(val); } }

/**************** Error Function *************/

void
IFT_SetError(char *fmt, ...) {
  va_list ap;
  va_start(ap,fmt);

  if (fmt != NULL) {
    vsprintf(IFT_Error,fmt,ap);
  }
  else {
    IFT_Error[0]='\0';
  }

  printf("IFT Error: %s\n",IFT_Error);
}

/*********************** pixbuf handling **************************/

static unsigned int
next_pow_2(unsigned int n) {
  int result = 8 * sizeof(unsigned int) - 1;

  while ( result && !(n & (1UL << result)) ) 
    result --;
  if ( !(n & ((1UL << result) - 1)) )
    return n;
  else 
    return 1UL << (result + 1);
}

IFT_Pixbuf * 
IFT_PixbufNew(unsigned int w, unsigned int h, unsigned int nchans, unsigned char *col) {
  IFT_Pixbuf * buf;
  int i;

  if (w<1 || w>100000) {
    IFT_SetError("Bad Pixbuf width");
    return NULL;
  }
  if (h<1 || h>100000) {
    IFT_SetError("Bad Pixbuf height");
    return NULL;
  }
  if (nchans <3 || nchans>4) {
    IFT_SetError("Bad Pixbuf channels");
    return NULL;
  }

  buf=(IFT_Pixbuf *)malloc(sizeof(IFT_Pixbuf));
  if (!buf) {
    IFT_SetError("1: Unable to allocate pixbuf");
    return NULL;
  }

  memset(buf,0,sizeof(IFT_Pixbuf));

  buf->aw = next_pow_2(w);
  buf->ah = next_pow_2(h);
  buf->w = w;
  buf->h = h;
  buf->rowlen = buf->aw * nchans;
  buf->nchans = nchans;

/*   printf("allocating %dx%d buffer for %dx%d image\n", */
/* 	 buf->aw, buf->ah, buf->w, buf->h); */

  buf->pixels = malloc(buf->aw * buf->ah * nchans);
  if (! buf->pixels) {
    IFT_SetError("2: Unable to allocate pixbuf");
    IFT_PixbufFree(buf);
    return NULL;
  }

  /* clear buffer*/
  for(i = 0; i < buf->w; i++) {
    *(buf->pixels+nchans*i+0) = col[0];
    *(buf->pixels+nchans*i+1) = col[1];
    *(buf->pixels+nchans*i+2) = col[2];
    if (nchans==4) 
      *(buf->pixels+nchans*i+3) = col[3];
  }
  for (i=1;i<buf->h;i++)
    memcpy(buf->pixels + i * buf->rowlen, buf->pixels, buf->w * nchans);

  return buf;
}

void
IFT_PixbufFree(IFT_Pixbuf * buf) {
  if (buf) {
    if (buf->pixels)
      free(buf->pixels);
    free(buf);
  }
  return ;
}

/***********************************************************************/

int
IFT_Init(char **fontpath) {
  int err;

  if (!fontpath) {
    IFT_SetError("NULL fontpath");
	 return 1;
  }

  err = FT_Init_FreeType( &Library );
  if (err) {
    IFT_SetError("Unable to initialize Freetype");
    return 1;
  }

  err = _IFT_FontListNew(fontpath);
  if (err) {
    IFT_SetError("Unable to initialize font list");
    return 1;
  }
  return 0;
}

int
IFT_Exit(void) {
  int error;
  error = FT_Done_FreeType( Library );
  if (error) {
    IFT_SetError("FT_Done_FreeType() failed");
    return 1;
  }

  _IFT_FontListFree();

  return 0;
}

IFT *
IFT_New(char * font,char * style) {
  IFT * ift;
  int err;

  ift=(IFT *)malloc(sizeof(IFT));
  if (!ift) {
    IFT_SetError("Unable to allocate IFT struct");
    return NULL;
  }
  memset(ift,0,sizeof(IFT));

  err = FT_New_Face( Library, _IFT_GetFontFile(font,style), 0,&(ift->face));
  if (err) {
    if (err == FT_Err_Unknown_File_Format) 
      IFT_SetError("Unknown face format");
    else 
      IFT_SetError("Unknown face error");

    IFT_Free(ift);
    return NULL;
  }

  ift->bg[0]=0;
  ift->bg[1]=0;
  ift->bg[2]=0;
  ift->bg[3]=255;

  ift->fg[0]=255;
  ift->fg[1]=255;
  ift->fg[2]=255;
  ift->fg[3]=255;

  ift->charsize=10;
  ift->border = 0;

  return ift;
}

void
IFT_Free(IFT * ift) {
  int i;
  if (ift) {
    if (FT_Done_Face(ift->face)) 
      IFT_SetError("Error: FT_Done_Face() failed");
    if (ift->glyphs) {
		 for (i=0;i<ift->num_glyphs;i++) {
			 if (ift->glyphs[i]) {
				 FT_Done_Glyph(ift->glyphs[i]);
			 }
		}
      free(ift->glyphs);
    }
    if (ift->pos){
      free(ift->pos);
    }
    if (ift->text) {
      free(ift->text);
    }
    if (ift->pixbuf) {
      IFT_PixbufFree(ift->pixbuf);
    }
    free(ift);
  }
  return;

}

int
IFT_SetBgColor(IFT * ift,
	       unsigned char r, unsigned char g,
	       unsigned char b, unsigned char a) {
  if (!ift) {
    IFT_SetError("Error: Bad IFT pointer");
    return 1;
  }

  ift->bg[0] = r;
  ift->bg[1] = g;
  ift->bg[2] = b;
  ift->bg[3] = a;
  return 0;
}

int
IFT_SetFgColor(IFT * ift,
	       unsigned char r, unsigned char g,
	       unsigned char b, unsigned char a) {
  if (!ift) {
    IFT_SetError("Error: Bad IFT pointer");
    return 1;
  }

  ift->fg[0] = r;
  ift->fg[1] = g;
  ift->fg[2] = b;
  ift->fg[3] = a;
  return 0;
}

int
IFT_SetSize(IFT * ift, unsigned int size, unsigned int border) {
  if (!ift) {
    IFT_SetError("Error: Bad IFT pointer");
    return 1;
  }

  ift->charsize=size;
  ift->border=border;
  return 0;
}

int
IFT_SetText(IFT * ift, char * text) {
  if (!ift) {
    IFT_SetError("Error: Bad IFT pointer");
    return 1;
  }

  ift->text=strdup(text);
  if (ift->text == NULL) 
    return 1;
  ift->textlen=strlen(ift->text);
  return 0;
}

int
IFT_Render(IFT * ift) {
  if (!ift) {
    IFT_SetError("Error: Bad IFT pointer");
    return 1;
  }

  if (_IFT_LoadGlyphs(ift)) {
    return 1;
  }

  if (_IFT_RenderToPixbuf(ift)) {
    return 1;
  }
  return 0;
}

/************************* static funcs ************************/

void
_IFT_Blit(IFT_Pixbuf* buf, FT_Bitmap bitmap, int x0, int y0, unsigned char *col) {
  int h,w,x,y,x2;
  unsigned char * src;
  unsigned char * dest;

  w=bitmap.width;
  h=bitmap.rows;
#ifdef _DEBUG
  printf("Blit: x,y=%d,%d w,h=%d,%d bw,bh=%d,%d\n", x0,y0,w,h,buf->w,buf->h);
#endif

  if (x0<0)
    x0=0;
  if (y0<0)
    y0=0;
  if (x0>buf->w)
    x0=buf->w;
  if (y0>buf->h)
    y0=buf->h;

  if (x0+w>buf->w)
    w=buf->w-x0;
  if (y0+h>buf->h)
    h=buf->h-y0;

#define colcomp(b,c,a) ((a*c + (255-a)*b)/255)
/* #define colcomp(b,c,a) (a == 0 ? 0 : c) */

  src=bitmap.buffer;
  dest=buf->pixels + y0*buf->rowlen + x0*buf->nchans;
  for (y=0;y<h;y++) {
    for (x=0;x<w;x++) {
      x2=x*buf->nchans;
      dest[x2+0]=colcomp(dest[x2+0], col[0], src[x]);
      dest[x2+1]=colcomp(dest[x2+1], col[1], src[x]);
      dest[x2+2]=colcomp(dest[x2+2], col[2], src[x]);
      if (buf->nchans == 4) 
        dest[x2+3]=colcomp(dest[x2+3], col[3], src[x]);
    }
    dest+=buf->rowlen;
    src+=bitmap.pitch;
  }
}

void
_IFT_CopyPixbuf(IFT_Pixbuf * dstbuf, int dstx, int dsty, IFT_Pixbuf * srcbuf,
                int srcx, int srcy, int srcw, int srch) {
  unsigned char * src;
  unsigned char * dst;
  int x,y,sx,dx;

  if (!srcbuf)
    return;
  if (!dstbuf)
    return;

  if (dstx<0)
    dstx=0;
  if (dsty<0)
    dsty=0;
  if (dstx>dstbuf->w)
    dstx=dstbuf->w;
  if (dsty>dstbuf->h)
    dsty=dstbuf->h;

  if (srcx<0)
    srcx=0;
  if (srcy<0)
    srcy=0;
  if (srcx>srcbuf->w)
    srcx=srcbuf->w;
  if (srcy>srcbuf->h)
    srcy=srcbuf->h;

  if (srcw<0)
    srcw=0;
  if (srch<0)
    srch=0;
  if (srcx+srcw>srcbuf->w)
    srcw=srcbuf->w-srcx;
  if (srcy+srch>srcbuf->h)
    srcy=srcbuf->h-srcy;

  src=srcbuf->pixels + srcx*srcbuf->nchans + srcy*srcbuf->rowlen;
  dst=dstbuf->pixels + dstx*dstbuf->nchans + (dsty + srch - 1)*dstbuf->rowlen;
  for (y=0;y<srch;y++) {
    for (x=0;x<srcw;x++) {
      sx=x*srcbuf->nchans;
      dx=x*dstbuf->nchans;
      dst[dx+0]=src[sx+0];
      dst[dx+1]=src[sx+1];
      dst[dx+2]=src[sx+2];
      if (dstbuf->nchans == 4) {
        if (srcbuf->nchans == 4)
          dst[dx+3]=src[sx+3];
        else 
          dst[dx+3]=255;
      }
    }
    src+=srcbuf->rowlen;
    dst-=dstbuf->rowlen;
  }
  return;
}

int
_IFT_LoadGlyphs (IFT * ift) {
  int err;
  FT_GlyphSlot  slot;
  FT_UInt       glyph_index;
  FT_UInt       previous;
  int pen_x, pen_y, n;
  int use_kerning;

  err = FT_Set_Char_Size(
                         ift->face,    /* handle to face object           */
                         0,       /* char_width in 1/64th of points  */
                         ift->charsize*64,   /* char_height in 1/64th of points */
                         100,     /* horizontal device resolution    */
                         100);    /* vertical device resolution      */

  if (err) {
    IFT_SetError("FT_Set_Char_Size() failed");
    return 1;
  }

  slot = ift->face->glyph;

  pen_x = 0;
  pen_y = 0;

  ift->glyphs=(FT_Glyph *)malloc(sizeof(FT_Glyph)*ift->textlen);
  if (!ift->glyphs) {
    IFT_SetError("glyphs alloc failed");
    return 1;
  }
  memset(ift->glyphs,0,sizeof(FT_Glyph)*ift->textlen);

  ift->pos=(FT_Vector *)malloc(sizeof(FT_Vector)*ift->textlen);
  if (!ift->pos) {
    IFT_SetError("glyphs pos failed");
    return 1;
  }
  memset(ift->pos,0,sizeof(FT_Vector)*ift->textlen);
  ift->num_glyphs=ift->textlen;

  use_kerning = FT_HAS_KERNING(ift->face);
  previous    = 0;

  for ( n = 0; n < ift->textlen; n++ ) {

    glyph_index = FT_Get_Char_Index( ift->face, ift->text[n] );

    if ( use_kerning && previous && glyph_index ) {
      FT_Vector  delta;
      FT_Get_Kerning( ift->face, previous, glyph_index,
                      ft_kerning_default, &delta );
      pen_x += delta.x >> 6;
    }

    ift->pos[ n ].x = pen_x;
    ift->pos[ n ].y = pen_y;

    err = FT_Load_Glyph( ift->face, glyph_index, FT_LOAD_DEFAULT );
    if (err) continue;  // ignore errors

    err = FT_Get_Glyph( ift->face->glyph, & (ift->glyphs[n]) );
    if (err) continue;  // ignore errors, jump to next glyph

    pen_x += slot->advance.x >> 6;
    previous = glyph_index;
  }

  return 0;
}

int
_IFT_RenderToPixbuf(IFT *ift) {
  FT_BBox  bbox;
  FT_BBox  glyph_bbox;
  int error,n;
  int width,height;
  int bufwidth,bufheight;
  int offx,offy;
  IFT_Pixbuf * bigbuf;
  int xmin,xmax,ymin,ymax;

  bbox.xMin = bbox.yMin =  32000;
  bbox.xMax = bbox.yMax = -32000;
  for ( n = 0; n < ift->num_glyphs; n++ ) {
    FT_Glyph_Get_CBox(ift->glyphs[n],ft_glyph_bbox_pixels, &glyph_bbox );
    glyph_bbox.xMin += ift->pos[n].x;
    glyph_bbox.xMax += ift->pos[n].x;
    glyph_bbox.yMin += ift->pos[n].y;
    glyph_bbox.yMax += ift->pos[n].y;

#ifdef _DEBUG
    printf("%c: bbox=%d=>%d,%d=>%d\n",ift->text[n],
              (int)glyph_bbox.xMin,
              (int)glyph_bbox.xMax,
              (int)glyph_bbox.yMin,
              (int)glyph_bbox.yMax);
#endif

    affect_if_min(bbox.xMin,glyph_bbox.xMin);
    affect_if_max(bbox.xMax,glyph_bbox.xMax);

    affect_if_min(bbox.yMin,glyph_bbox.yMin);
    affect_if_max(bbox.yMax,glyph_bbox.yMax);
  }

  if ( bbox.xMin > bbox.xMax ) {
    bbox.xMin = 0;
    bbox.yMin = 0;
    bbox.xMax = 0;
    bbox.yMax = 0;
  }

  width = (bbox.xMax - bbox.xMin);
  height = (bbox.yMax - bbox.yMin);
  bufwidth=width+2*ift->border;
  bufheight=2*height;
  offx = (( bufwidth  - width )/2)*64;
  offy = (( bufheight - height)/2)*64;
  bigbuf=IFT_PixbufNew(bufwidth,bufheight,4,ift->bg);
  if (!bigbuf) {
    IFT_SetError("3: Unable to allocate pixbuf (%dx%d)", bufwidth,bufheight);
    return 1;
  }

#ifdef _DEBUG
  printf("String: %dx%d border=%d charsz=%d\n",width,height,ift->border,ift->charsize);
  printf("Bigbuf: %dx%d\n",bufwidth,bufheight);
#endif
  xmin=bufwidth;
  xmax=0;
  ymin=bufheight;
  ymax=0;
  /* blit in a BIG pixbuf and calculte *REAL* bounding box */
  for ( n = 0; n < ift->num_glyphs; n++ ) {
    FT_Glyph  image;
    FT_Vector pen;

    image = ift->glyphs[n];
    pen.x = offx + 64*ift->pos[n].x;
    pen.y = offy + 64*ift->pos[n].y;
    error = FT_Glyph_To_Bitmap( &image, ft_render_mode_normal, 0, 0 );
    if (!error) {
      FT_BitmapGlyph bitmap = (FT_BitmapGlyph)image;

      _IFT_Blit(bigbuf,
                bitmap->bitmap,
                bitmap->left + pen.x / 64,
                bufheight - bitmap->top - pen.y / 64,
                ift->fg);

      if (bitmap->bitmap.width != 0 &&
          bitmap->bitmap.rows !=0) {
        affect_if_min(xmin,bitmap->left + pen.x / 64);
        affect_if_max(xmax,bitmap->left+bitmap->bitmap.width + pen.x / 64);
        affect_if_min(ymin,bufheight-bitmap->top - pen.y / 64);
        affect_if_max(ymax,bufheight-bitmap->top + bitmap->bitmap.rows - pen.y / 64);
      }

      FT_Done_Glyph( image );
    }
  }


#ifdef _DEBUG
  printf("BBox: %d,%d => %d,%d (%dx%d)\n",
            xmin,ymin,xmax,ymax,xmax-xmin,ymax-ymin);
#endif

  bufwidth=xmax-xmin+2*ift->border;
  bufheight=ymax-ymin+2*ift->border;

#ifdef _DEBUG
  printf("Pixbuf: %dx%d offset=%dx%d\n",bufwidth,bufheight,offx/64,offy/64);
#endif
        
  ift->pixbuf=IFT_PixbufNew(bufwidth,bufheight,4,ift->bg);
  if (!ift->pixbuf) {
    IFT_SetError("4: Unable to allocate pixbuf");
    IFT_PixbufFree(bigbuf);
    return 1;
  }
        
  _IFT_CopyPixbuf(
                  ift->pixbuf,
                  ift->border,
                  ift->border,
                  bigbuf,
                  xmin,
                  ymin,
                  xmax-xmin,
                  ymax-ymin);
                
  IFT_PixbufFree(bigbuf);

  return 0;
}

int
_IFT_FontListNew(char **dirnames) {
  char ** dir;

  if (!dirnames)
	  return 1;

  _IFT_FontListFree();

  /* lookup in dirs */
  dir=dirnames;
  while (*dir) {
#ifdef _DEBUG
    printf("Searching in dir %s\n",*dir);
#endif

	 _IFT_FontLookup(*dir);
    dir++;
  }

  if (FontListNum == 0)
    return 1;

  qsort(
        FontList,
        FontListNum,
        sizeof(IFT_FontRec),
        _IFT_FontRecCmpName
        );

  return 0;
}

int
_IFT_FontRecCmpName(const void *p1, const void *p2) {
  IFT_FontRec * f1, *f2;
  int ret=-1;

  f1=(IFT_FontRec *)p1;
  f2=(IFT_FontRec *)p2;

  if (f1->name && f2->name) {
    return strcasecmp(f1->name,f2->name);
  }

  return ret;
}

int
_IFT_FontRecCmpFile(const void *p1, const void *p2) {
  IFT_FontRec * f1, *f2;
  int ret=-1;

  f1=(IFT_FontRec *)p1;
  f2=(IFT_FontRec *)p2;

  if (f1->file && f2->file) {
    return strcasecmp(f1->file,f2->file);
  }

  return ret;
}

void
_IFT_FontListFree(void) {
  int i;
  if (FontList) {
    for (i=0;i<FontListNum;i++) {
      if (FontList[i].name)
        free(FontList[i].name);
      if (FontList[i].file)
        free(FontList[i].file);
      if (FontList[i].fullpath)
        free(FontList[i].fullpath);
    }
    free(FontList);
    FontList=NULL;
  }
  FontListNum=0;
}

int
_IFT_FontLookup(char *dirname) {
  DIR * dir;
  struct dirent *entry;
  char filename[1024];
  struct stat file;

#ifdef _DEBUG
  printf("Opening dir %s\n",dirname);
#endif

  dir = opendir(dirname);
  if (!dir) {
#ifdef _DEBUG
  printf("Error while opening dir %s...Skipping.\n",dirname);
#endif
    return 1;
  }
        
  while ((entry=readdir(dir))) {
    if (!strcmp(entry->d_name,"."))
      continue;
    if (!strcmp(entry->d_name,".."))
      continue;

    sprintf(filename,"%s/%s",dirname,entry->d_name);

    if (stat(filename,&file))
      continue;

    if (S_ISDIR(file.st_mode)) { /* recursive search */
      _IFT_FontLookup(filename);
    }
    else if (S_ISREG(file.st_mode)) { /* add font to file */
      _IFT_FontAdd(entry->d_name,filename);
    }
  }

  closedir(dir);

  return 0;
}

int
_IFT_FontAdd(char * file, char * fullpath) {
  FT_Face face;
  IFT_FontRec * newflist;
  char fontname[512];

  if (FT_New_Face(Library,fullpath,0,&face))
    return 1;

  if (! (face->face_flags & FT_FACE_FLAG_SCALABLE))
    return 1;

  FontListNum++;

  newflist=realloc(FontList,FontListNum*sizeof(IFT_FontRec));
  if (!newflist) 
    return 1;

  FontList=newflist;

  sprintf(fontname,"%s:%s",face->family_name,face->style_name);

  FontList[FontListNum-1].name=strdup(fontname);
  FontList[FontListNum-1].file=strdup(file);
  FontList[FontListNum-1].fullpath=strdup(fullpath);

  FT_Done_Face(face);
  return 0;
}

char *
_IFT_GetFontFile(char * name,char * style) {
  char * res;
  char altname[512];

  if (style && strcmp(style,"")) 
	  sprintf(altname,"%s:%s",name,style);
  else
	  sprintf(altname,"%s",name);

  res=_IFT_SearchFontFile(altname,NULL);
  if (res)
    return res;

  sprintf(altname,"%s:Regular",name);
  res=_IFT_SearchFontFile(altname,NULL);
  if (res)
    return res;

  res=_IFT_SearchFontFile(NULL,name);
  if (res)
    return res;

  return "";
}

char *
_IFT_SearchFontFile(char * name, char * file) {
  IFT_FontRec * font,target;
  int i;

#ifdef _DEBUG
  if (name)
    printf("Looking for name=%s\n",name);
  if (file)
    printf("Looking for file=%s\n",file);
#endif

  target.name=name;
  target.file=file;
  target.fullpath=NULL;

  /* binary search in sorted array */
  font=bsearch(&target,FontList,FontListNum,sizeof(IFT_FontRec),_IFT_FontRecCmpName);
  if (font) {
#ifdef _DEBUG
    printf("Found %s => %s\n",font->name,font->fullpath);
#endif
    return font->fullpath;
  }

  /* linear search in non sorted array */
  for(i=0;i<FontListNum;i++) {
    font=&FontList[i];
    if (!_IFT_FontRecCmpFile(font,&target)) {
#ifdef _DEBUG
      printf("Found %s => %s\n",font->name,font->fullpath);
#endif
      return font->fullpath;
    }
  }

#ifdef _DEBUG
  printf("Not found\n");
#endif

  return NULL;
}

