/*
 * $Id: ih-core.c,v 1.1.1.1 2002/03/25 13:07:26 blah Exp $
 */
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <GL/gl.h>
#include "svg-parser.h"

#define WITH_IH_DIVX
#define WITH_IH_PIXBUF
/* #define WITH_IH_TEXT */

#ifdef WITH_IH_DIVX
#include "ih-divx.h"
#endif /* WITH_IH_DIVX */

#ifdef WITH_IH_PIXBUF
#include "ih-pixbuf.h"
#endif /* WITH_IH_PIXBUF */

#ifdef WITH_IH_TEXT
#include "ih-text.h"
#endif /* WITH_IH_TEXT */


/* some static structs */

#ifdef WITH_IH_DIVX
static char* 
divx_suffixes[] = { ".avi", ".divx", 0 };

#endif /* WITH_IH_DIVX */

#ifdef WITH_IH_PIXBUF
static char* pixbuf_suffixes[] = { ".gif", ".jpeg", ".jpg",
                                   ".png", 0 };

#endif /* WITH_IH_PIXBUF */

/**
 * image_handlers_def:
 * @type:
 * @suffixes: zero-terminated array of managed file name suffixes
 **/
static struct {
  struct ih_handler handler;
  char *type;
  char **suffixes;
}
image_handlers_def[] = {

#ifdef WITH_IH_DIVX
  /* HANDLER_DIVX */
  {
    {
      divx_init,                /* init */
      divx_exit,                /* exit */
      divx_new,                 /* new */
      divx_destroy,             /* destroy */
      divx_decodeframe,         /* decodeframe */
      divx_sync,		/* sync frame */
      NULL,                     /* seek */
      divx_seekstart,           /* seek start */
      NULL,                     /* miss */
      divx_releasefb,           /* releasefb */
      divx_open,                /* open */
      divx_close,               /* close */
      divx_debug,               /* debug */
    },
    "divx",                     /* type */
    divx_suffixes               /* suffixes */
  },
#endif /* WITH_IH_DIVX */

#ifdef WITH_IH_PIXBUF
  /* HANDLER_PIXBUF */
  {
    {
      NULL,                     /* init */
      NULL,                     /* exit */
      pixbuf_new,               /* new */
      pixbuf_destroy,           /* destroy */
      pixbuf_decodeframe,       /* decodeframe */
      pixbuf_sync,		/* sync frame */
      NULL,                     /* seek */
      pixbuf_seekstart,         /* seek start */
      NULL,                     /* miss */
      pixbuf_releasefb,         /* releasefb */
      pixbuf_open,              /* open */
      pixbuf_close,             /* close */
      pixbuf_debug,             /* debug */
    },
    "pixbuf",                   /* type */
    pixbuf_suffixes             /* suffixes */
  },
#endif /* WITH_IH_PIXBUF */

#ifdef WITH_IH_TEXT
  /* HANDLER_TEXT */
  {
    {
      text_init,                /* init */
      text_exit,                /* exit */
      text_new,                 /* new */
      text_destroy,             /* destroy */
      text_decodeframe,         /* decodeframe */
      text_sync,		/* sync frame */
      NULL,                     /* seek */
      text_seekstart,           /* seek start */
      NULL,                     /* miss */
      text_releasefb,           /* releasefb */
      text_open,                /* open */
      text_close,               /* close */
      text_debug,               /* debug */
    },
    "text",                     /* type */
    NULL                        /* suffixes */
  },
#endif /* IH_TEXT */
  { { 0 } }			/* end of array */
};

/* module management */

/**
 * ih_init:
 * init all modules.
 *
 * returns: 1 on success, 0 on error.
 **/
ih_boolean_t
ih_init(void) {
  int typeid;

  for (typeid = 0; image_handlers_def[typeid].type; typeid ++) {
    if (image_handlers_def[typeid].handler.init)
      if (!image_handlers_def[typeid].handler.init())
	return 0;
  }

  return 1;
}

/**
 * ih_exit:
 *
 * exit image handler.
 **/
void
ih_exit(void) {
  int typeid;

  for (typeid = 0; image_handlers_def[typeid].type; typeid ++) {
    if (image_handlers_def[typeid].handler.exit)
      image_handlers_def[typeid].handler.exit();
  }

  return;
}

ih_context_t
ih_context_new(void * userdata) {
  ih_context_t ctx;

  ctx=(ih_context_t)malloc(sizeof(struct ih_context));
  if (!ctx)
    return NULL;

  ctx->userdata=userdata;

  return ctx;
}


void ih_context_destroy(ih_context_t ctx) 
{
  if (ctx) free(ctx);
}

/* guess handler from filename suffix */
static struct ih_handler *
get_type_handler(dtdtc_string_t name) {
  char *dot;
  int i, typeid;
  dot = strrchr(name, '.');

  for (typeid = 0; image_handlers_def[typeid].suffixes; typeid ++) {
    for (i = 0;
	 image_handlers_def[typeid].suffixes[i];
	 i++) 
      if (!strcasecmp(dot, image_handlers_def[typeid].suffixes[i])) {
	return &image_handlers_def[typeid].handler;
    }    
  }
  return NULL;
}

/* instance management */

/**
 * ih_image_new:
 *
 * handler frontend.
 *
 * returns image handler on success, NULL on error.
 **/
ih_image_t
ih_image_new(svg_element_image_t image, dtdtc_string_t name) {
  ih_image_t ih;

  ih = calloc(1, sizeof(*ih));
  if (!ih) return NULL;

  /* lookup handler */
  ih->handler = get_type_handler(name);
  if (!ih->handler) {
    free(ih);
    return NULL; /* unknown file type */
  }
  ih->image = image;

  /* allocate derived class members */
  ih = ih->handler->new(ih);
  if (!ih) return NULL;

  if ( !ih->handler->open(ih) ) {
    free(ih);
    return NULL;                /* failed to open file */
  }
  
  return ih;
}

/* free image */
void ih_image_destroy(ih_context_t ctx, ih_image_t ih) {
  if (ctx && ih && ih->handler) 
    ih->handler->destroy(ctx,ih);
}


/* virtual methods */

ih_boolean_t
ih_image_decodeframe(ih_image_t ih, ih_boolean_t render_p, ih_boolean_t *updated) {
  return ih->handler->decodeframe(ih, render_p, updated);
}

ih_boolean_t
ih_image_seekstart(ih_image_t ih, ih_boolean_t render_p) {
  return ih->handler->seekstart(ih, render_p);
}

ih_boolean_t
ih_image_releasefb(ih_image_t ih) {
  return ih->handler->releasefb(ih,ih->frameno);
}

ih_boolean_t
ih_image_sync(ih_image_t ih, struct timeval *now, struct timeval *start, ih_boolean_t render_p, ih_boolean_t *firstsync) {
  return ih->handler->sync(ih,now,start,render_p,firstsync);
}

/* accessors */

int
ih_image_width(ih_image_t ih) {
  return ih->width;
}

int
ih_image_height(ih_image_t ih) {
  return ih->height;
}

int
ih_image_alignment(ih_image_t ih) {
  return ih->alignment;
}


int
ih_image_frameno(ih_image_t ih) {
  return ih->frameno;
}

int
ih_image_frames(ih_image_t ih) {
  return ih->frames;
}

void *
ih_image_pixels(ih_image_t ih) {
  return ih->pixels;
}

int
ih_image_channels(ih_image_t ih) {
  return ih->channels;
}

int
ih_image_pixformat(ih_image_t ih) {
  return ih->pixformat;
}

float
ih_image_x_max(ih_image_t ih) {
  return ih->x_max;
}

float
ih_image_y_max(ih_image_t ih) {
  return ih->y_max;
}

/* utility */
unsigned int ih_next_pow_2(unsigned int n) {
  int result = 8 * sizeof(unsigned int) - 1;

  while ( result && !(n & (1UL << result)) ) 
    result --;
  if ( !(n & ((1UL << result) - 1)) )
    return n;
  else 
    return 1UL << (result + 1);
}
