#include "svg-parser.h"
#include <string.h>
#include <ctype.h>

/* a chunk is a run of consecutive segments */
typedef struct svg_path_chunk *svg_path_chunk_t;
struct svg_path_chunk {
  GLint op;			/* GL_LINE, GL_LINE_LOOP etc... */
  int n_points;			/* number of points is chunk */
  float *points;		/* point coordinates */
  svg_path_chunk_t next;	/* next chunk */
};


typedef enum {
  PATHSEG_UNKNOWN                      = 0,
  PATHSEG_CLOSEPATH                    = 1,
  PATHSEG_MOVETO_ABS                   = 2,
  PATHSEG_MOVETO_REL                   = 3,
  PATHSEG_LINETO_ABS                   = 4,
  PATHSEG_LINETO_REL                   = 5,
  PATHSEG_CURVETO_CUBIC_ABS            = 6,
  PATHSEG_CURVETO_CUBIC_REL            = 7,
  PATHSEG_CURVETO_QUADRATIC_ABS        = 8,
  PATHSEG_CURVETO_QUADRATIC_REL        = 9,
  PATHSEG_ARC_ABS                      = 10,
  PATHSEG_ARC_REL                      = 11,
  PATHSEG_LINETO_HORIZONTAL_ABS        = 12,
  PATHSEG_LINETO_HORIZONTAL_REL        = 13,
  PATHSEG_LINETO_VERTICAL_ABS          = 14,
  PATHSEG_LINETO_VERTICAL_REL          = 15,
  PATHSEG_CURVETO_CUBIC_SMOOTH_ABS     = 16,
  PATHSEG_CURVETO_CUBIC_SMOOTH_REL     = 17,
  PATHSEG_CURVETO_QUADRATIC_SMOOTH_ABS = 18,
  PATHSEG_CURVETO_QUADRATIC_SMOOTH_REL = 19,
  PATHSEG_LINE_STRIP,
  PATHSEG_LINE_LOOP
} svg_path_seg_type_t;


/**
 * pump parameters for command `type' 
 *
 * returns 1 on success, 0 on error.
 */
static int
svg_path_next_param(svg_path_seg_type_t type, 
		    dtdtc_string_t *current, float params[7]) {
  dtdtc_string_t next[7];

  switch( type ) {
  case PATHSEG_UNKNOWN:
    return 0;
  case PATHSEG_MOVETO_ABS:
  case PATHSEG_MOVETO_REL:
  case PATHSEG_LINETO_ABS:
  case PATHSEG_LINETO_REL:
  case PATHSEG_CURVETO_QUADRATIC_SMOOTH_ABS:
  case PATHSEG_CURVETO_QUADRATIC_SMOOTH_REL:
    /* (x y) */
    params[0] = strtof(*current, &next[0]);
    if (next[0] != *current) {
      params[1] = strtof(next[0], &next[1]);
      if (next[1] != next[0])
	*current = next[1];
	return 1;
    }
    break;
  case PATHSEG_CLOSEPATH:
    /* (none) */
    return 1;
  case PATHSEG_LINETO_HORIZONTAL_ABS:
  case PATHSEG_LINETO_HORIZONTAL_REL:
  case PATHSEG_LINETO_VERTICAL_ABS:
  case PATHSEG_LINETO_VERTICAL_REL:
    /* (x) */
    params[0] = strtof(*current, &next[0]);
    if (next[0] != *current) {
      *current = next[0];
      return 1;
    }
    break;
  case PATHSEG_ARC_ABS:
  case PATHSEG_ARC_REL:
    /* (rx ry x-axis-rotation large-arc-flag sweep-flag x y) */
    params[0] = strtof(*current, &next[0]);
    if (next[0] != *current) {
      params[1] = strtof(next[0], &next[1]);
      if (next[1] != next[0]) {
	params[2] = strtof(next[1], &next[2]);
	if (next[2] != next[1]) {
	  params[3] = strtof(next[2], &next[3]);
	  if (next[3] != next[2]) {
	    params[4] = strtof(next[3], &next[4]);
	    if (next[4] != next[3]) {
	      params[5] = strtof(next[4], &next[5]);
	      if (next[5] != next[4]) {
		params[6] = strtof(next[5], &next[6]);
		if (next[6] != next[5]) {
		  params[7] = strtof(next[6], &next[7]);
		  if (next[7] != next[6]) {
		    *current = next[7];
		    return 1;
		  }}}}}}}}
    break;
  case PATHSEG_CURVETO_CUBIC_ABS:
  case PATHSEG_CURVETO_CUBIC_REL:
  case PATHSEG_CURVETO_QUADRATIC_ABS:
  case PATHSEG_CURVETO_QUADRATIC_REL:
  case PATHSEG_CURVETO_CUBIC_SMOOTH_ABS:
  case PATHSEG_CURVETO_CUBIC_SMOOTH_REL:
  default:
  }
  return 0;
}

static svg_path_seg_type_t
svg_path_next_token(dtdtc_string_t *parse) {
  dtdtc_string_t current = *parse;
  svg_path_seg_type_t type;

  while ( isspace(*current) ) current ++;

  switch (*current) {
  case '\0': current = NULL;
    *parse = NULL;
    return PATHSEG_UNKNOWN;
  case 'm': 
    type = PATHSEG_MOVETO_REL; break;
  case 'M':
    type = PATHSEG_MOVETO_ABS; break;
  case 'l': 
    type = PATHSEG_LINETO_REL; break;
  case 'L':
    type = PATHSEG_LINETO_ABS; break;
  case 'h': 
    type = PATHSEG_LINETO_HORIZONTAL_REL; break;
  case 'H':
    type = PATHSEG_LINETO_HORIZONTAL_ABS; break;
  case 'v': 
    type = PATHSEG_LINETO_VERTICAL_REL; break;
  case 'V':
    type = PATHSEG_LINETO_VERTICAL_ABS; break;
  case 'z': 
  case 'Z':
    type = PATHSEG_CLOSEPATH; break;
  default:
    type = PATHSEG_UNKNOWN; break;
  }
  
  do {
    current ++;
  } while ( isspace(*current) );
  *parse = current;
  return type;
}


static void
svg_path_build_lists(svg_path_data_t path_data,
		     svg_path_chunk_t chunk) {
  int i;
  svg_path_chunk_t root = chunk;

  path_data->fill_id = glGenLists(2);
  path_data->stroke_id = path_data->fill_id + 1;

  glNewList(path_data->fill_id, GL_COMPILE);
  while (chunk) {
    glBegin(GL_TRIANGLE_STRIP);
    for (i=0; i<chunk->n_points; i++) {
      glVertex2f(chunk->points[2 * i], -chunk->points[2 * i + 1]);
    }
    glEnd();
    chunk = chunk->next;
  }
  glEndList();

  chunk = root;

  glNewList(path_data->stroke_id, GL_COMPILE);
  while (chunk) {
    glBegin(GL_LINE_STRIP);
    for (i=0; i<chunk->n_points; i++) {
      glVertex2f(chunk->points[2 * i], -chunk->points[2 * i + 1]);
    }
    glEnd();
    chunk = chunk->next;
  }
  glEndList();
}

static void
print_path(svg_path_chunk_t chunk) {
  int i;
  while (chunk) {
    printf("\n%s:",
	   chunk->op == PATHSEG_LINE_STRIP ? "strip" : 
	   chunk->op == PATHSEG_LINE_LOOP ? "loop" : "???");
    for (i=0; i<chunk->n_points; i++) {
      printf("->(%f, %f) ", 
	     chunk->points[2*i],
	     chunk->points[2*i + 1]);
    }
    chunk = chunk->next;
  }
}

static svg_path_chunk_t
svg_path_chunk_new(svg_path_seg_type_t op) {
  svg_path_chunk_t chunk;
  chunk = malloc( sizeof(struct svg_path_chunk) );
  if (!chunk) return NULL;
  chunk->next=NULL;
  chunk->n_points = 0;
  chunk->points = (float*)NULL;
  chunk->op = op;
  return chunk;
}

static void
svg_path_chunk_destroy(svg_path_chunk_t chunk) {
  svg_path_chunk_t next;

  while (chunk) {
    next = chunk->next;
    if (chunk->points) free(chunk->points);
    free(chunk);
    chunk = next;
  }
}

static void
svg_path_chunk_add_vertex(svg_path_chunk_t chunk, float x, float y) {
  chunk->n_points ++;
  if (chunk->points)
    chunk->points = realloc( chunk->points,
			     chunk->n_points * 2 * sizeof(float) );
  else {
    chunk->points = malloc( 2 * sizeof(float) );
  }
  if (!chunk->points) return;
  chunk->points[2*(chunk->n_points - 1)] = x;
  chunk->points[2*(chunk->n_points - 1) + 1] = y;
}

#define ENTER_DRAW_LINE do {				\
  if (!in_draw) {					\
    svg_path_chunk_t _chunk;				\
    if (prev_type != PATHSEG_CLOSEPATH) {		\
      sub_x = pen_x;					\
      sub_y = pen_y;					\
    }							\
							\
    in_draw = 1;					\
    _chunk = svg_path_chunk_new(PATHSEG_LINE_STRIP);	\
    if (!root)						\
      root = _chunk;					\
    if (!_chunk) {					\
      svg_path_chunk_destroy(root);			\
      return NULL;					\
    }							\
    _chunk->op = PATHSEG_LINE_STRIP;			\
    svg_path_chunk_add_vertex(_chunk, pen_x, pen_y);	\
    if (chunk) chunk->next = _chunk;			\
    chunk = _chunk;					\
  }							\
} while (0)

#define LEAVE_DRAW do {				\
  sub_x = pen_x;				\
  sub_y = pen_x;				\
  in_draw = 0;					\
} while (0)

static svg_path_chunk_t
svg_path_compute_chunks(svg_path_data_t path_data,
			const dtdtc_string_t value) {
  svg_path_seg_type_t type, prev_type = PATHSEG_UNKNOWN;
  svg_path_chunk_t root = NULL, chunk = NULL;
  dtdtc_string_t parse = value;
  float params[7] = {0};
  float pen_x = 0, pen_y = 0;
  float sub_x = 0, sub_y = 0;
  int in_draw = 0;

  do {
    type = svg_path_next_token(&parse);
    if (parse) {
      switch( type ) {

      case PATHSEG_UNKNOWN:
	LEAVE_DRAW;
	break;

      case PATHSEG_MOVETO_ABS:
	while ( svg_path_next_param(type, &parse, params) ) {
	  pen_x = params[0];
	  pen_y = params[1];      
	}
	LEAVE_DRAW;
	break;

      case PATHSEG_MOVETO_REL:
	while ( svg_path_next_param(type, &parse, params) ) {
	  pen_x += params[0];
	  pen_y += params[1];
	}
	LEAVE_DRAW;
	break;

      case PATHSEG_CLOSEPATH:
	if (chunk)
	  svg_path_chunk_add_vertex(chunk, sub_x, sub_y);
	break;

      case PATHSEG_LINETO_ABS:
	ENTER_DRAW_LINE;
	while ( svg_path_next_param(type, &parse, params) ) {
	  pen_x = params[0];
	  pen_y = params[1];
	  svg_path_chunk_add_vertex(chunk, pen_x, pen_y);
	}
	break;

      case PATHSEG_LINETO_REL:
	ENTER_DRAW_LINE;
	while ( svg_path_next_param(type, &parse, params) ) {
	  pen_x += params[0];
	  pen_y += params[1];
	  svg_path_chunk_add_vertex(chunk, pen_x, pen_y);
	}
	break;

      case PATHSEG_LINETO_HORIZONTAL_ABS:
	ENTER_DRAW_LINE;
	while ( svg_path_next_param(type, &parse, params) ) {
	  pen_x = params[0];
	  svg_path_chunk_add_vertex(chunk, pen_x, pen_y);
	}
	break;

      case PATHSEG_LINETO_HORIZONTAL_REL:
	ENTER_DRAW_LINE;
	while ( svg_path_next_param(type, &parse, params) ) {
	  pen_x += params[0];
	  svg_path_chunk_add_vertex(chunk, pen_x, pen_y);
	}
	break;

      case PATHSEG_LINETO_VERTICAL_ABS:
	ENTER_DRAW_LINE;
	while ( svg_path_next_param(type, &parse, params) ) {
	  pen_y = params[0];
	  svg_path_chunk_add_vertex(chunk, pen_x, pen_y);
	}
	break;

      case PATHSEG_LINETO_VERTICAL_REL:
	ENTER_DRAW_LINE;
	while ( svg_path_next_param(type, &parse, params) ) {
	  pen_y += params[0];
	  svg_path_chunk_add_vertex(chunk, pen_x, pen_y);
	}
	break;

      case PATHSEG_CURVETO_CUBIC_ABS:
      case PATHSEG_CURVETO_CUBIC_REL:
      case PATHSEG_CURVETO_QUADRATIC_ABS:
      case PATHSEG_CURVETO_QUADRATIC_REL:
      case PATHSEG_ARC_ABS:
      case PATHSEG_ARC_REL:
      case PATHSEG_CURVETO_CUBIC_SMOOTH_ABS:
      case PATHSEG_CURVETO_CUBIC_SMOOTH_REL:
      case PATHSEG_CURVETO_QUADRATIC_SMOOTH_ABS:
      case PATHSEG_CURVETO_QUADRATIC_SMOOTH_REL:
      default:
      }
    }
    prev_type = type;
  } while (parse);

  return root;
  
}

void
svg_path_data_set(dtdtc_context_t ctx,
		  svg_path_data_t path_data,
		  const dtdtc_string_t value, void *data) {
  svg_path_chunk_t chunk;

  chunk = svg_path_compute_chunks(path_data, value);
  svg_path_build_lists(path_data, chunk);
/*   print_path(chunk); */
  svg_path_chunk_destroy(chunk);
}

void
svg_path_end(dtdtc_context_t ctx) {
  struct svg_element_path *path;

  dtdtc_debug(&ctx->error, "");

#if 0
  svg_player_t player = (svg_player_t)ctx->user_data;
#endif

  path = (struct svg_element_path*)ctx->element;

  dtdtc_debug(&ctx->error, "%fx%f+%f+%f",
	      path->width, path->height, path->x, path->y);

  if (path->fill.paintType == SVG_PAINTTYPE_UNKNOWN)
    dtdtc_attribute_inherit(ctx, ctx->element, "fill",
			    sizeof(struct svg_paint_server));

  if (path->stroke.paintType == SVG_PAINTTYPE_UNKNOWN)
    dtdtc_attribute_inherit(ctx, ctx->element, "stroke",
			    sizeof(struct svg_paint_server) );

}

void
svg_path_draw(svg_element_path_t path) {

/*   printf("path stroke-width=%f\n", path->stroke_width); */

  glDisable(GL_TEXTURE_2D);

  glPushMatrix();
  glScalef(1.0 / 450, 1.0 / 450, 1.0);

  /* fill */
#if 1
  if (path->fill.paintType == SVG_PAINTTYPE_RGBCOLOR) {
    glColor4f( path->fill.color.r/255.0, path->fill.color.g/255.0,
	       path->fill.color.b/255.0, path->fill.color.a/255.0 );
    glCallList(path->d.fill_id);  
  }
#endif
  /* draw border */
  if (path->stroke.paintType == SVG_PAINTTYPE_RGBCOLOR) {
    glColor4f( path->stroke.color.r/255.0, path->stroke.color.g/255.0,
	       path->stroke.color.b/255.0, path->stroke.color.a/255.0 );
/*     glColor4f(1.0, 1.0, 1.0, 1.0 ); */
    glLineWidth(path->stroke_width);

    glCallList(path->d.stroke_id);

  }

    glCallList(path->d.stroke_id);
  glPopMatrix();
}

void
svg_path_destroy(dtdtc_context_t ctx) {
  dtdtc_debug(&ctx->error, "");
}
